;;;; tabular-list.scm
;;;; Kon Lovett, May '07

;; Issues
;;
;; - Use format for justification & fill - headers, labels, & cells.
;; Per item & per category.
;;
;; - Have borders - top/bottom/left/right outside/inside, inner (this is spacing).
;;
;; - Have easy api for building justification & fill format-string.
;;
;; - Have easy api for building borders.
;;
;; - Have pre-defined sets of list-tabular specifications.
#|
       Header
       ------ ...
-----
Label  Cell   ...
  .     .
  .     .
  .     .
|#

(use srfi-1)
(use miscmacros misc-extn-list)

(eval-when (compile)
  (declare
    (not usual-integrations
      format
      string-length)
    (fixnum)
    (inline)
    (no-procedure-checks)
    (no-bound-checks)
    (import
      format)
    (bound-to-procedure
      format)
    (export
      list-tabular-spacing
      list-tabular-formatting
      list-tabular-compel
      %display-list-tabular!
      *display-list-tabular!
      display-list-tabular!
      display-list-tabular) ) )
;;;

(define-constant DEFAULT-HEADER '())
(define-constant DEFAULT-JUSTIFICATION '((0 . "") (1 . "")))
(define-constant DEFAULT-SPACING 2)
(define DEFAULT-TO-STRING ->string)

;;;

(define (spacing? obj)
  (and (fixnum? obj) (< 0 obj)) )

(define (formatting-specification? obj)
  (list? obj) )

;;;

(define-parameter list-tabular-spacing DEFAULT-SPACING
  (lambda (x)
    (if (spacing? x)
      x
      (list-tabular-spacing))))

(define-parameter list-tabular-formatting DEFAULT-JUSTIFICATION
  (lambda (x)
    (if (formatting-specification? x)
      x
      (list-tabular-formatting))))

(define-parameter list-tabular-compel DEFAULT-TO-STRING
  (lambda (x)
    (if (procedure? x)
      x
      (list-tabular-compel))))

;;

(define (display-fill-char chr #!optional (cnt 1))
  (repeat cnt (display chr)) )

; 
; 

(define (%display-list-tabular! rowlst hdr fmtlst btwn to-string)
  ;
  (let* ([has-hdr (not-null? hdr)]
         [rows (if has-hdr (cons hdr rowlst) rowlst)])
    ;
    (define (stringify-rows)
      (map!
        (lambda (cols)
          (map!
            (lambda (col)
              (let ([str (to-string col)])
                (cons (string-length str) str) ) )
            cols) )
        rows) )
    ;
    (let ([strows (stringify-rows)])

      (define (max-col-wids)
        (fold
          (lambda (cols wids)
            (map!
              (lambda (wid col)
                (fxmax (car col) wid) )
              wids cols) )
          (make-list (length (first strows)) 0)
          strows) )
      ;
      (let ([wids (max-col-wids)])

        (define (disp-row cols)
          (for-each
            (lambda (col wid)
              (display (cdr col))
              (display-fill-char #\space (fx+ btwn (fx- wid (car col)))) )
            cols wids)
          (newline) )
        ;
        (define (disp-bars)
          (for-each
            (lambda (wid)
              (display-fill-char #\- wid)
              (display-fill-char #\space btwn) )
            wids)
          (newline) )
        ;
        (when has-hdr
          (disp-row (first strows))
          (disp-bars)
          (set! strows (cdr strows)) )
        ;
        (for-each (cut disp-row <>) strows) ) ) ) )

;;

(define (*display-list-tabular! rowlst hdr fmtlst btwn to-string)
  ;
  (unless (and (pair? rowlst) (list? (first rowlst)))
    (error 'display-list-tabular "invalid row list" rowlst) )
  (unless (and (fixnum? btwn) (fx< 0 btwn))
    (error 'display-list-tabular "invalid spacing" btwn) )
  (unless (procedure? to-string)
    (error 'display-list-tabular "invalid procedure" to-string) )
  ;
  (%display-list-tabular! rowlst hdr fmtlst btwn to-string) )

;;

(define (display-list-tabular! rowlst #!key (header DEFAULT-HEADER) (label DEFAULT-LABEL) (formatting (list-tabular-formatting)) (spacing (list-tabular-spacing)) (compel (list-tabular-compel)))
  (*display-list-tabular! rowlst header formatting spacing compel) )

;;

(define (display-list-tabular rows #!key (header DEFAULT-HEADER) (label DEFAULT-LABEL) (formatting (list-tabular-formatting)) (spacing (list-tabular-spacing)) (compel (list-tabular-compel)))
  (*display-list-tabular!
    (map (cut list-copy <>) rows)
    (list-copy header)
    formatting spacing compel) )
